{
                 CPU16ASM - Assembler for CPU16

    (C) Copyright 2008-2009 John B. Stephensen  All Rights Reserved

  This software in source, object or binary form is licensed only for
  personal non-profit educational use in the Amateur Radio Service and
  the license is not transferrable. The software is provided as-is for
  experimental purposes and the author does not warranty its freedom
  from defects or its suitability for any specific application.

  This assembler processes an ASCII text file containing CPU16 assembly
  language statements and produces a Verilog file for Xilinx ISE 10.1i that is
  combined with the CPU source files to generate the CPU. The assembler can
  create files with one 16,384 x 18 block RAM, two 2,048 x 9 block RAMs or four
  4,096 x 4 block RAMs that contain the machine code corresponding to the
  assembly language program. The Verilog module is named "PROM". The number of
  RAMs to use is specified by the PRAM command and defaults to 2. Machine code
  consists of 16-bit instructions for the processor described in "A Soft
  Processor for DSP" in the Winter 2009 edition of the TAPR Packet Status
  Register. There are 87 instructions and 5 compiler directives. The source
  file must have the extension ".asm". In addition to the Verilog file, the
  assembler produces a text file containing source code and the corresponding
  machine code in hexidecimal form.

  Assembly language statements are one line each with with an optional label
  field, followed by a mnuemonic instruction and one or more operand fields.
  Labels are alphanumeric strings at the beginning of a line. They specify the
  locations of instructions in the program memory. Instructions are
  alphanumeric strings preceeded by a blank space. Operands follow instructions
  and are alphanumeric strings specifying a data value, an instruction address
  or reserved values (such as R0-R15) that specify registers in the processor.
  They may also be numeric strings that specify constant data values or
  addresses in program or data memory. A string beginning with "$" is in
  hexidecimal format and a string beginning with "#" is in binary format. "+",
  "-" or anything beginning with 0-9 specifies decimal format. Operands are
  separated by commas and labels are followed by colons.

  History:
    8-28-08 update instruction set
    3-16-09 fixed problem with 4 RAM option and added 1 RAM option
    4-29-09 generate UCF file instead of Verilog file
}
program CPU16ASM;
{$APPTYPE CONSOLE}

uses
  SysUtils;

type
  ptype = (ok,error); {parser output}
  symtype = (nothing,colon,comma,number,name,operation,regnum,unknown);
  symtv = record {symbol type and value}
    t: symtype;
    v: integer
  end;
  instype = record {instruction mnuemonic, encoding and type of parameters}
    n: string[4];
    e: integer;
    t: (equ,regdef,ds,org,d8,d16,a9,a14,p0,r1,r1a,r2,r1d4,r1d4a,r2d4,r1d7,r1d8,r1d8n,r1d16,r1a8l,r1a8h,rom);
  end;

{global constants}
const
  maxsym: integer = 999; {maximum number of symbols in symbol table}
  maxreg: integer = 15; {register names}
  regtbl: array [0..15] of string[3] =
    ('R0','R1','R2','R3','R4','R5','R6','R7','R8','R9',
     'R10','R11','R12','R13','R14','R15');
  maxrw: integer = 120; {reserved words}
  rwtbl: array [0..120] of instype =
    ((n:'LDL';e:$8000;t:r1d8),   {load registers}
     (n:'LDH';e:$9000;t:r1d8),
     (n:'LOAD';e:$8000;t:r1d16), {LDL (here) followed by LDH (previous entry)}
     (n:' ';e:$9000;t:r1a8h),
     (n:'LDA';e:$8000;t:r1a8l),  {LDL (here) followed by LDH (previous entry)}
     (n:'EQU';e:0;t:equ),        {name a constant}
     (n:'REG';e:0;t:regdef),     {name a register}
     (n:'DS';e:0;t:ds),          {reserve 16-bit memory locations}
     (n:'ORG';e:0;t:org),        {set program counter}
     (n:'PRAM';e:0;t:rom),       {set number of program RAMs}
     (n:'DW';e:0;t:d16),         {define word}
     (n:'CALL';e:$0000;t:a14),   {call}
     (n:'JMP';e:$4000;t:a9),     {jump}
     (n:'JL';e:$4400;t:a9),
     (n:'JT';e:$4400;t:a9),
     (n:'JNL';e:$4600;t:a9),
     (n:'JF';e:$4600;t:a9),
     (n:'JZ';e:$4800;t:a9),
     (n:'JNZ';e:$4A00;t:a9),
     (n:'JC';e:$4C00;t:a9),
     (n:'JAGE';e:$4C00;t:a9),
     (n:'JBLE';e:$4C00;t:a9),
     (n:'JNC';e:$4E00;t:a9),
     (n:'JALT';e:$4E00;t:a9),
     (n:'JBGT';e:$4E00;t:a9),
     (n:'RET';e:$5000;t:p0),     {ret}
     (n:'RL';e:$5400;t:p0),
     (n:'RNL';e:$5600;t:p0),
     (n:'RZ';e:$5800;t:p0),
     (n:'RNZ';e:$5A00;t:p0),
     (n:'RC';e:$5C00;t:p0),
     (n:'RAGE';e:$5C00;t:p0),
     (n:'RBLE';e:$5C00;t:p0),
     (n:'RNC';e:$5E00;t:p0),
     (n:'RALT';e:$5E00;t:p0),
     (n:'RBGT';e:$5E00;t:p0),
     (n:'LOOP';e:$5100;t:p0),    {loop}
     (n:'LL';e:$5500;t:p0),
     (n:'LNL';e:$5700;t:p0),
     (n:'LZ';e:$5900;t:p0),
     (n:'LNZ';e:$5B00;t:p0),
     (n:'LC';e:$5D00;t:p0),
     (n:'LAGE';e:$5D00;t:p0),
     (n:'LBLE';e:$5D00;t:p0),
     (n:'LNC';e:$5F00;t:p0),
     (n:'LALT';e:$5F00;t:p0),
     (n:'LBGT';e:$5F00;t:p0),
     (n:'MARK';e:$6000;t:p0),    {mark beginning of loop}
     (n:'REP';e:$6800;t:d8),     {repeat}
     (n:'UMUL';e:$A000;t:r2),    {unsigned multiply and optional accumulate}
     (n:'UMLN';e:$A100;t:r2),
     (n:'UMAC';e:$A200;t:r2),
     (n:'UMSB';e:$A300;t:r2),
     (n:'MUL';e:$A400;t:r2),    {signed multiply and optional accumulate}
     (n:'MULN';e:$A500;t:r2),
     (n:'MAC';e:$A600;t:r2),
     (n:'MSUB';e:$A700;t:r2),
     (n:'FDIV';e:$A800;t:p0),    {continue division to ger fractional result}
     (n:'IDIV';e:$A900;t:r2),    {divide Ra by Rb}
     (n:'TST';e:$AC00;t:r1d4),   {test bit - place value in link register}
     (n:'CLL';e:$AD00;t:p0),     {clear link bit}
     (n:'STL';e:$AD10;t:p0),     {set link bit}
     (n:'CLC';e:$AE00;t:p0),     {clear carry bit}
     (n:'STC';e:$AE10;t:p0),     {set carry bit}
     (n:'CMP';e:$B300;t:r2),     {compare}
     (n:'CMPC';e:$B500;t:r2),    {same with carry propagation}
     (n:'CPI';e:$B700;t:r1d4),   {compare immediate (0-15)}
     (n:'CPZ';e:$B700;t:r1),     {compare to zero}
     (n:'OUT';e:$B800;t:r1d7),   {output to port}
     (n:'IN';e:$C000;t:r1d7),    {input from port}
     (n:'WR';e:$C900;t:r2),      {write to memory}
     (n:'RD';e:$CA00;t:r2),      {read memory}
     (n:'XCHG';e:$CB00;t:r2),    {read then write memory}
     (n:'ADZ';e:$D000;t:r2),     {register math - alter Z and C}
     (n:'NEG';e:$D100;t:r2),
     (n:'ADD';e:$D200;t:r2),
     (n:'SUB';e:$D300;t:r2),
     (n:'ADC';e:$D400;t:r2),     {register math with carry propogation}
     (n:'SBC';e:$D500;t:r2),
     (n:'ADI';e:$D600;t:r1d4),   {4-bit immediate data}
     (n:'SUI';e:$D700;t:r1d4),
     (n:'INC';e:$D610;t:r1),
     (n:'DEC';e:$D710;t:r1),
     (n:'MOV';e:$D800;t:r2),     {copy 16-bit register (scalar or complex)}
     (n:'CNEG';e:$DB00;t:r2),    {complex register math}
     (n:'CADD';e:$DC00;t:r2),
     (n:'CSUB';e:$DF00;t:r2),
     (n:'NOT';e:$E000;t:r2),     {2 register logic operations}
     (n:'AND';e:$E100;t:r2),
     (n:'OR';e:$E200;t:r2),
     (n:'NOP';e:$E200;t:p0),     {R0 or R0}
     (n:'XOR';e:$E300;t:r2),
     (n:'CLR';e:$E300;t:r1a),    {XOR register with itself}
     (n:'MASK';e:$E400;t:r1d4),  {zero upper bits}
     (n:'RST';e:$E500;t:r1d4),   {bit manipulation}
     (n:'SET';e:$E600;t:r1d4),
     (n:'INV';e:$E700;t:r1d4),
     (n:'MLLS';e:$E800;t:r2),    {move bytes and extend sign}
     (n:'MHLS';e:$E900;t:r2),
     (n:'SXL';e:$E800;t:r1a),    {extend sign}
     (n:'SXH';e:$E900;t:r1a),
     (n:'INS4';e:$EA00;t:r2),    {rotate right and insert 4, 8 or 12 bits}
     (n:'INS8';e:$EB00;t:r2),
     (n:'IN12';e:$EC00;t:r2),
     (n:'ROR4';e:$EA00;t:r1a),   {rotate right by 4, 8 or 12 bits}
     (n:'SWAP';e:$EB00;t:r1a),
     (n:'ROL4';e:$EC00;t:r1a),
     (n:'REV';e:$ED00;t:r2),     {reverse bit order}
     (n:'SHR';e:$EE00;t:r1d4),   {shift with arbitrary end bits}
     (n:'SHL';e:$EF00;t:r1d4),
     (n:'ROR';e:$EE00;t:r1),     {specific useful shifts}
     (n:'ROL';e:$EF50;t:r1),
     (n:'RRC';e:$EE80;t:r1),
     (n:'RLC';e:$EF90;t:r1),
     (n:'LSR';e:$EEC0;t:r1),
     (n:'LSL';e:$EFD0;t:r1),
     (n:'ASR';e:$EE40;t:r1),
     (n:'LPL';e:$F000;t:r1),     {retreive multiply/divide results}
     (n:'LPH';e:$F100;t:r1),
     (n:'LQ';e:$F200;t:r1),
     (n:'LR';e:$F300;t:r1)
     );

{global variables}
var
  f,f2: TextFile; {currently open files}
  fn: string[127]; {file name}
  errors: integer; {error counter}
  maxaddr: integer; {maximum memory address}
  romsize: integer; {number of ROMs to place code in}
  buf: string[132]; {line buffer}
  pos: integer; {current position in line buffer}
  len: integer; {length of line in buffer}
  line: integer; {line number}
  paddr: integer; {current instruction and data address}
  symnam: array[0..999] of string[32]; {table of symbols in assembly file}
  symval: array[0..999] of integer; {symbol value}
  symtyp: array[0..999] of (undef,reg,addrp,addrd,data); {symbol type}
  lastsym: integer; {current size of table}
  addr: array[0..4095] of integer; {instruction table index for each line no.}
  inst: array[0..4095] of integer; {number of instructions for each line no.}
  op0: array[0..4095] of integer; {instruction}
  op1: array[0..4095] of integer; {register A}
  op2: array[0..4095] of integer; {register B or 4-9-12-bit constant}
  op3: array[0..4095] of integer; {4-bit constant}
{
  Report error and increment error counter
}
procedure report(s: string);
begin
  errors := errors + 1;
  writeln('*** ',s,' on line ',line,'!');
  write('Press ENTER to continue...'); readln;
end; {report}
{
  Print symbol type and value
}
procedure printsym (s: symtv);
begin
  {write('  ');
  case s.t of
  nothing: write('nothing');
  colon: write('colon');
  comma: write('comma');
  number: write('number');
  name: write('name');
  operation: write('operation');
  regnum: write('register');
  unknown: write('unknown');
  end;
  write(':',s.v);
  if s.t = name
  then writeln('=',symnam[s.v])
  else writeln;}
end; {printsym}
{
 Decode symbols in line buffer and return type and value.
}
function getsym: symtv;
var
  i: integer;  {index into line buffer}
  n: integer; {numeric value}
  s: string[32]; {alphanumeric value}
begin
  {default return values}
  getsym.t := nothing;
  getsym.v := 0;
  {return if end of buffer reached}
  if (pos > len) then exit;
  {skip white space}
  while (buf[pos] <= ' ') and (pos <= len) do pos := pos+1;
  {return if end of buffer reached}
  if (pos > len) then exit;
  {a semicolon indicates that the rest of the line is a comment}
  if buf[pos] = ';' then
  begin
    pos := len; {ignore rest of line}
    exit;
  end; {if}
  {check for a comma}
  if buf[pos] = ',' then
  begin
    pos := pos+1;
    getsym.t := comma;
    exit;
  end; {if}
  {check for a colon}
  if buf[pos] = ':' then
  begin
    pos := pos+1;
    getsym.t := colon;
    exit;
  end; {if}
  {check for numeric value}
  if ((buf[pos] = '+') or (buf[pos] = '-') or (buf[pos] = '$') or
    (buf[pos] = '#') or ((buf[pos] >= '0') and (buf[pos] <= '9')))
    and (pos <= len) then
  begin
    if buf[pos] = '#' then
    begin
      {process binary number}
      pos := pos+1;
      {accumulate value}
      n := 0;
      while ((buf[pos] = '0') or (buf[pos] = '1')) and (pos <= len) do
      begin
        n := n*2;
        if (buf[pos] = '1') then n := n + 1;
        pos := pos+1;
      end; {do}
      {and return value}
      getsym.t := number;
      getsym.v := n;
      exit;
    end; {if}
    if buf[pos] = '$' then
    begin
      {process hexidecimal number}
      pos := pos+1;
      {accumulate value}
      n := 0;
      while (((buf[pos] >= '0') and (buf[pos] <= '9')) or
        ((buf[pos] >= 'A') and (buf[pos] <= 'F')) or
        ((buf[pos] >= 'a') and (buf[pos] <= 'f'))) and (pos <= len) do
      begin
        n := n*16;
        if ((buf[pos] >= '0') and (buf[pos] <= '9'))
        then n := n + (integer(buf[pos]) - integer('0'));
        if ((buf[pos] >= 'A') and (buf[pos] <= 'F'))
        then n := n + (integer(buf[pos]) - integer('A') + 10);
        if ((buf[pos] >= 'a') and (buf[pos] <= 'f'))
        then n := n + (integer(buf[pos]) - integer('a') + 10);
        pos := pos+1;
      end; {do}
      {and return value}
      getsym.t := number;
      getsym.v := n;
      exit;
    end; {if}
    i := pos; {save sign}
    {skip sign}
    if (buf[pos] = '+') or (buf[pos] = '-') then pos := pos+1;
    {accumulate absolute value}
    n := 0;
    while ((buf[pos] >= '0') and (buf[pos] <= '9') and (pos <= len)) do
    begin
      n := n*10;
      n := n + (integer(buf[pos]) - integer('0'));
      pos := pos+1;
    end; {do}
    {negate if necessary}
    if buf[i] = '-' then n := -n;
    {and return value}
    getsym.t := number;
    getsym.v := n;
    exit;
  end; {if}
  {check for alphanumeric string beginning with letter}
  if (((buf[pos] >= 'A') and (buf[pos] <= 'Z')) or
    ((buf[pos] >= 'a') and (buf[pos] <= 'z'))) and (pos <= len) then
  begin
    i := 0;
    {it may contain numbers or letters}
    while (((buf[pos] >= '0') and (buf[pos] <= '9')) or
      ((buf[pos] >= 'A') and (buf[pos] <= 'Z')) or
      ((buf[pos] >= 'a') and (buf[pos] <= 'z'))) and (pos <= len) do
    begin
      if i < 32 then i := i+1;
      s[i] := buf[pos];
      pos := pos+1;
    end; {while}
    s[0] := char(i);
  end else begin
    {error if not alphanumeric string}
    getsym.t := unknown;
    report('Syntax error');
    exit;
  end; {if}
  {check for register name}
  for i := 0 to maxreg do
  begin
    if s = regtbl[i] then
    begin
      {return index if found}
      getsym.t := regnum;
      getsym.v := i;
      exit;
    end; {if}
  end; {for}
  {check for reserved words}
  for i := 0 to maxrw do
  begin
    if s = rwtbl[i].n then
    begin
      {return index if found}
      getsym.t := operation;
      getsym.v := i;
      exit;
    end; {if}
  end; {for}
  {check if previously defined symbol}
  if lastsym > 0 then
  begin
    {at least 1 entry present so scan table}
    for i := 1 to lastsym do
    begin
      if s = symnam[i] then
      begin
        {done if matching entry found - return index}
        getsym.t := name;
        getsym.v := i;
        exit;
      end; {if}
    end; {for}
  end; {if}
  {not defined so create new entry at end of table}
  if lastsym < maxsym then
  begin
    lastsym := lastsym+1; {increase length}
    symnam[lastsym] := s; {name}
    symval[lastsym] := 0; {undefined value}
    symtyp[lastsym] := undef; {undefined type}
  end else begin
    report('Symbol table overflow');
  end;
  {and return index}
  getsym.t := name;
  getsym.v := lastsym;
end; {getsym}
{
  Get index of address that may already be defined or could be defined later
}
function getaddress9: integer;
var
  s: symtv; {word type and value}
begin
  getaddress9 := -1; {default}
  {get next symbol from line buffer}
  s := getsym;
  printsym(s);
  if s.t = name
  then getaddress9 := s.v {index of address in symval}
  else report('Relative address missing');
end; {getaddress9}
function getaddress14: integer;
var
  s: symtv; {word type and value}
begin
  getaddress14 := -1; {default}
  {get next symbol from line buffer}
  s := getsym;
  printsym(s);
  if s.t = name
  then getaddress14 := s.v {index of address in symval}
  else report('Absolute address missing');
end; {getaddress14}
{
  Get constant - may be number or symbol
}
function getcons: integer;
var
  s: symtv; {word type and value}
begin
  getcons := 0; {default return value}
  {get next symbol from line buffer}
  s := getsym;
  printsym(s);
  {check symbol type}
  case s.t of
  number: if (s.v >= 0) and (s.v <= 15)
          then getcons := s.v {numeric value}
          else report('Constant not 0-15');
  name: if (symtyp[s.v] = data) and (symval[s.v] >= 0) and (symval[s.v] <= 15)
        then getcons := symval[s.v]
        else report('Undefined symbol or wrong type or value not 0-15');
  else report('Missing constant');
  end; {case}
  {check range}
end; {getcons}
{
  Get port number - may be number or symbol
}
function getport: integer;
var
  s: symtv; {word type and value}
begin
  getport := 0; {default return value}
  {get next symbol from line buffer}
  s := getsym;
  printsym(s);
  {check symbol type}
  case s.t of
  number: if (s.v >= 0) and (s.v <= 127)
          then getport := s.v {numeric value}
          else report('Port number not 0-127');
  name: if (symtyp[s.v] = data) and (symval[s.v] >= 0) and (symval[s.v] <= 127)
        then getport := symval[s.v]
        else report('Undefined symbol or wrong type or value not 0-127');
  else report('Missing constant or port number');
  end; {case}
  {check range}
end; {getport}
{
  Get immediate data parameter - may be number or symbol
}
function getdata8: integer;
var
  s: symtv; {word type and value}
begin
  getdata8 := 0; {default return value}
  {get next symbol from line buffer}
  s := getsym;
  printsym(s);
  {check symbol type}
  case s.t of
  number: if (s.v >= -128) and (s.v <= 255)
          then getdata8 := s.v and 255 {numeric value}
          else report('Immediate data not betwwen -128 and +255');
  name: if (symtyp[s.v] = data) or (symtyp[s.v] = addrd)
        then getdata8 := symval[s.v] and 255
        else report('Undefined symbol or wrong type of symbol');
  else report('Missing 8-bit data');
  end; {case}
end; {getdata8}
function getdata16: integer;
var
  s: symtv; {word type and value}
begin
  getdata16 := 0; {default return value}
  {get next symbol from line buffer}
  s := getsym;
  printsym(s);
  {check symbol type}
  case s.t of
  number: if (s.v >= -32768) and (s.v <= 65535)
          then getdata16 := s.v and 65535 {numeric value}
          else report('Immediate data not betwwen -32,768 and +65,536');
  name: if (symtyp[s.v] = data) or (symtyp[s.v] = addrd)
        then getdata16 := symval[s.v] and 65535
        else report('Undefined symbol or wrong type of symbol');
  else report('Missing 16-bit data');
  end; {case}
end; {getdata16}
{
  eat comma
}
procedure eatcomma;
var
  s: symtv; {word type and value}
begin
  {get next symbol from line buffer}
  s := getsym;
  printsym(s);
  {which must be comma}
  if s.t <> comma
  then report('Missing comma');
end;
{
  Get register number - may be R0-R15 or symbol with value 0-15
}
function getreg: integer;
var
  s: symtv; {word type and value}
begin
  getreg := 0; {default}
  {get next symbol from line buffer}
  s := getsym;
  printsym(s);
  {and check type}
  case s.t of
  regnum: getreg := s.v; {return register number}
  name: if (symtyp[s.v] = reg)
        then getreg := symval[s.v] {or return numeric value of symbol}
        else report('Undefined symbol or wrong type or value out of range');
  else report('Invalid register'); {error}
  end; {case}
end;
{
 Process line buffer by decoding each field and creating instruction entry
}
procedure parseline;
var
  i: integer;
  s: symtv; {word type and value}
  n: integer; {label location in table}
  r: integer; {register number}
  d: integer; {immediate data}
begin
  {check instruction address and return if error}
  if paddr > maxaddr then
  begin
    report('Program memory exhausted');
    writeln(' paddr = ',paddr,' maxaddr = ',maxaddr,' romsize = ',romsize);
    exit;
  end;
  {set operation type and operand to invalid values}
  addr[line] := -1;
  op0[paddr] := -1;
  op1[paddr] := -1;
  op2[paddr] := -1;
  {get first symbol}
  pos := 1;
  s := getsym;
  printsym(s);
  {check for label preceding operation}
  n := 0;
  if s.t = name then
  begin
    n := s.v; {save location in table}
    {check for colon}
    s := getsym;
    printsym(s);
    if s.t = colon then
    begin
      {check for duplicate definition}
      if symtyp[n] = undef {undefined address}
      then begin
        symval[n] := paddr; {set to current instruction address}
        symtyp[n] := addrp; {with program address type}
      end else begin report('Duplicate symbol'); exit; end; {duplicate}
      n := 0; {EQU cannot follow colon}
      {fetch symbol after colon}
      s := getsym;
      printsym(s);
    end; {if}
  end; {if}
  {check for op code}
  case s.t of
    nothing: exit; {label only - end of line}
    operation: i := s.v; {save index into reserved word table}
    else begin report('Invalid opcode'); exit; end; {error if not operation}
  end; {case}
  {check for psuedoinstructions}
  if rwtbl[i].t = equ then
  begin
    if n <= 0 then begin report('Missing label for EQU'); exit; end;
    if symtyp[n] = undef then
    begin
      {put constant in table}
      symval[n] := getdata16; {set value}
      symtyp[n] := data; {data}
    end else report('Duplicate symbol');
    exit;
  end; {if}
  if rwtbl[i].t = regdef then
  begin
    if n <= 0 then begin report('Missing label for REG'); exit; end;
    if symtyp[n] = undef then
    begin
      {put constant in table}
      symval[n] := getreg; {set value}
      symtyp[n] := reg; {register number}
    end else report('Duplicate symbol');
    exit;
  end; {if}
  if rwtbl[i].t = ds then
  begin
    if n <= 0 then begin report('Missing label for DS'); exit; end;
    if symtyp[n] = undef then
    begin
      {put data address in table and then increment data address}
      symval[n] := paddr; {set value}
      symtyp[n] := addrd; {type is data address}
      paddr := paddr + getdata16;
    end else report('Duplicate symbol');
    exit;
  end; {if}
  if rwtbl[i].t = org then
  begin
    {set data address}
    paddr := getdata16;
    exit;
  end; {if}
  if rwtbl[i].t = rom then
  begin
    {set data address}
    romsize := getdata8;
    if romsize = 0
    then maxaddr := 16383
    else maxaddr := romsize*1024-1;
    exit;
  end; {if}
  {process real instructions}
  addr[line] := paddr; {save instruction address}
  inst[line] := 1; {default to one generated instruction}
  op0[paddr] := i; {save index into reserved word table}
  {get operands}
  case rwtbl[i].t of
  d8: {8-bit data format}
    begin
      op1[paddr] := getdata8 - 2; {only used for REP instruction}
      op2[paddr] := 0;
      op3[paddr] := 0;
    end;
  a9: {relative address format}
    begin
      op1[paddr] := getaddress9;
      op2[paddr] := 0;
      op3[paddr] := 0;
    end;
  a14: {absolute address format}
    begin
      op1[paddr] := getaddress14;
      op2[paddr] := 0;
      op3[paddr] := 0;
    end;
  p0: {no operand format}
    begin
      op1[paddr] := 0;
      op2[paddr] := 0;
      op3[paddr] := 0;
    end;
  r1: {single register format}
    begin
      op1[paddr] := getreg;
      op2[paddr] := 0;
      op3[paddr] := 0;
    end;
  r1a: {single register format for two register instruction}
    begin
      r := getreg;
      op1[paddr] := r;
      op2[paddr] := r;
      op3[paddr] := 0;
    end;
  r2: {two register format}
    begin
      op1[paddr] := getreg;
      eatcomma;
      op2[paddr] := getreg;
      op3[paddr] := 0;
    end;
  r1d4: {single register plus 4-bit data format}
    begin
      op1[paddr] := getreg;
      eatcomma;
      op2[paddr] := getcons;
      op3[paddr] := 0;
    end;
  r1d4a: {single register plus 4-bit data format for two register instructions}
    begin
      r := getreg;
      op1[paddr] := r;
      op2[paddr] := r;
      eatcomma;
      op3[paddr] := getcons;
    end;
  r2d4: {two registers plus 4-bit data format}
    begin
      op1[paddr] := getreg;
      eatcomma;
      op2[paddr] := getreg;
      eatcomma;
      op3[paddr] := getcons;
    end;
  r1d7: {single register plus 7-bit port format}
    begin
      op1[paddr] := getreg;
      eatcomma;
      op2[paddr] := getport;
      op3[paddr] := 0;
    end;
  r1d8: {single register plus 8-bit data format}
    begin
      op1[paddr] := getreg;
      eatcomma;
      op2[paddr] := getdata8;
      op3[paddr] := 0;
    end;
  r1d8n: {single register plus negated 8-bit data format}
    begin
      op1[paddr] := getreg;
      eatcomma;
      op2[paddr] := (-getdata8) and 255;
      op3[paddr] := 0;
    end;
  r1d16: {single register plus 16-bit data format}
    begin
      {get parameter values}
      r := getreg;
      eatcomma;
      d := getdata16;
      {generate 1st instruction}
      op1[paddr] := r;
      op2[paddr] := d and 255;
      op3[paddr] := 0;
      {generate 2nd instruction}
      op0[paddr+1] := op0[paddr] - 1; {use previous symbol to get LDH}
      op1[paddr+1] := r;
      op2[paddr+1] := (d shr 8) and 255;
      op3[paddr+1] := 0;
      paddr := paddr+1;
      inst[line] := 2;
    end;
  r1a8l: {single register plus absolute address}
    begin
      {get parameter values}
      r := getreg;
      eatcomma;
      d := getaddress14;
      {generate 1st instruction}
      op1[paddr] := r;
      op2[paddr] := d;
      op3[paddr] := 0;
      {generate 2nd instruction}
      op0[paddr+1] := op0[paddr] - 1; {use previous symbol to get LDH}
      op1[paddr+1] := r;
      op2[paddr+1] := d;
      op3[paddr+1] := 0;
      paddr := paddr+1;
      inst[line] := 2;
    end;
  d16: {16-bit data format (DW)}
    begin
      {set instruction to parameter value}
      op1[paddr] := getdata16;
      op2[paddr] := 0;
      op3[paddr] := 0;
    end;
  end; {case}
  {increment instruction address}
  paddr := paddr+1;
end; {parseline}

{write hexidecimal number to screen}
procedure disphex (n: integer);
var
  c: char;
begin
  case n of
  0: c := '0';
  1: c := '1';
  2: c := '2';
  3: c := '3';
  4: c := '4';
  5: c := '5';
  6: c := '6';
  7: c := '7';
  8: c := '8';
  9: c := '9';
  10: c := 'A';
  11: c := 'B';
  12: c := 'C';
  13: c := 'D';
  14: c := 'E';
  15: c := 'F';
  else c := '0';
  end; {case}
  write(c);
end; {disphex}

{write hexidecimal number to file}
procedure writehex (n: integer);
var
  c: char;
begin
  case n of
  0: c := '0';
  1: c := '1';
  2: c := '2';
  3: c := '3';
  4: c := '4';
  5: c := '5';
  6: c := '6';
  7: c := '7';
  8: c := '8';
  9: c := '9';
  10: c := 'A';
  11: c := 'B';
  12: c := 'C';
  13: c := 'D';
  14: c := 'E';
  15: c := 'F';
  else c := '0';
  end; {case}
  write(f,c);
end; {writehex}

{pass 1 reads file f and generates machine code}
procedure pass1;
var
  i: integer; {loop index}
  a: integer; {address}
begin
  Reset(f);
  {process each line}
  while not eof(f) do
  begin
    readln(f,buf);
    len := length(buf);
    line := line+1;
    writeln(line,': ',buf,' (',len,')'); {echo line}
    if len > 0 then parseline;
    {write('Press ENTER to continue...'); readln;}
  end; {while}
  {close source file}
  CloseFile(f);
  {resolve addresses}
  for i := 0 to paddr-1 do
  begin
    {process only jump and call instructions}
    {relative addresses}
    if rwtbl[op0[i]].t = a9 then
    begin
      {report unresolved addresses}
      if symtyp[op1[i]] = undef then
      begin
        errors := errors+1;
        writeln('*** Relative address ',symnam[op1[i]],' is undefined!');
      end; {if}
      a := symval[op1[i]];
      if (a < 0) or (a > maxaddr) then
      begin
        errors := errors+1;
        writeln('*** Relative address ',symnam[op1[i]],' is too large!');
      end; {if}
      {fix relative address by subtracting address of next instruction}
      op1[i] := (a - i - 1) and 511; {and truncate to 8 bits}
      if ((a-i-1) < -256) or ((a-i-1) > 255) then
      begin
        errors := errors+1;
        writeln('*** Relative address ',symnam[op1[i]],' is too far from jump!');
      end; {if}
    end; {if}
    {absolute addresses}
    if rwtbl[op0[i]].t = a14 then
    begin
      {report unresolved addresses}
      if symtyp[op1[i]] = undef then
      begin
        errors := errors+1;
        writeln('*** Absolute address ',symnam[op1[i]],' is undefined!');
      end; {if}
      a := symval[op1[i]];
      if (a < 0) or (a > maxaddr) then
      begin
        errors := errors+1;
        writeln('*** Absolute address ',symnam[op1[i]],' is too large!');
      end; {if}
      {fix absolute address}
      op1[i] := a;
    end; {if}
    {lower half of absolute addresses}
    if rwtbl[op0[i]].t = r1a8l then
    begin
      {report unresolved addresses}
      if symtyp[op2[i]] = undef then
      begin
        errors := errors+1;
        writeln('*** Lower half of address ',symnam[op2[i]],' is undefined!');
      end; {if}
      a := symval[op2[i]];
      if (a < 0) or (a > maxaddr) then
      begin
        errors := errors+1;
        writeln('*** Lower half of address ',symnam[op2[i]],' is too large!');
      end; {if}
      {fix lower half of absolute address}
      op2[i] := a and 255;
    end; {if}
    {upper half of absolute addresses}
    if rwtbl[op0[i]].t = r1a8h then
    begin
      {report unresolved addresses}
      if symtyp[op2[i]] = undef then
      begin
        errors := errors+1;
        writeln('*** Upper half of address ',symnam[op2[i]],' is undefined!');
      end; {if}
      a := symval[op2[i]];
      if (a < 0) or (a > maxaddr) then
      begin
        errors := errors+1;
        writeln('*** Upper half of address ',symnam[op2[i]],' is too large!');
      end; {if}
      {fix upper half of absolute address}
      op2[i] := (a shr 8) and 255;
    end; {if}
  end; {for}
  {complete encoding}
  for a := 0 to paddr-1 do
  begin
    {writeln(op0[a],':',op1[a],':',op2[a],':',op3[a]);}
    op0[a] := rwtbl[op0[a]].e + op1[a] + (16*op2[a]) + (256*op3[a]);
  end; {for}
end; {pass1}

{display 16-bit integer in hexidecimal}
procedure write16(i:integer);
begin
    writehex((i shr 12) and 15);
    writehex((i shr 8) and 15);
    writehex((i shr 4) and 15);
    writehex(i and 15);
end; {write16}

{print machine language instruction for each line}
procedure decode(a:integer);
begin
  if a >= 0 then
  begin
    {display machine code}
    write16(a); {address}
    write(f,': ');
    write16(op0[a]); {instruction}
  end else begin
    {fill space}
    write(f,'          ');
  end; {if}
end; {decode}

{create interlisted source and object code}
procedure pass2;
var
  i: integer;
begin
  {initialize}
  i := 0;
  {open source file}
  AssignFile(f2,fn+'.asm');
  Reset(f2);
  if ioresult = 0 then
  begin
    {create file}
    AssignFile(f,fn+'.txt');
    Rewrite(f);
    if ioresult = 0 then
    begin
      {header}
      writeln(f,'CPU16 Assembler v0.9  (C) Copyright 2009 John B. Stephensen');
      writeln(f,errors,' errors, ',lastsym,' symbols used, ',paddr,' instructions.');
      writeln(f);
      writeln(f,'Line Addr  Inst Source');
      {process each line}
      while not eof(f2) do
      begin
        readln(f2,buf);
        i := i+1;
        write(f,i:4,' '); {line number}
        decode(addr[i]); {machine coded instruction}
        writeln(f,' ',buf); {echo line}
        if inst[i] > 1 then
        begin
          write(f,'     '); {no line number}
          decode(addr[i]+1); {possible second instruction}
          writeln(f); {no text}
        end; {if}
      end; {while}
      {close source file}
      CloseFile(f2);
      {add symbol table}
      writeln(f);
      writeln(f,'Sym  Type  Hex. Dec. Name');
      for i := 0 to lastsym do
      begin
        write(f,i:3,': ');
        case symtyp[i] of
        undef: write(f,'UNDEF ');
        reg: write(f,'REG   ');
        addrp: write(f,'PADDR ');
        addrd: write(f,'DADDR ');
        data: write(f,'DATA  ');
        end;
        write16(symval[i]); {display hex value then decimal value}
        writeln(f,' ',symval[i]:4,' ',symnam[i]);
      end; {for}
      CloseFile(f);
    end else begin
      writeln('ERROR creating listing file!');
    end; {if}
  end else begin
    writeln('ERROR opening source file on second pass!');
  end; {if}
end; {pass2}

{write word to file least significant digit first}
procedure writeword (n: integer);
begin
  writehex(n and 15);
  writehex((n shr 4) and 15);
  writehex((n shr 8) and 15);
  writehex((n shr 12) and 15);
end; {writeword}

{write out one init parameter for 16-bit memory}
{n is the initialization string number}
procedure init16(n: integer; data: array of integer);
var
  i: integer;
begin
  write(f,'INST "pgmram" INIT_');
  writehex((n shr 4) and 15);
  writehex(n and 15);
  write(f,' = 256''h');
  for i := 15 downto 0 do
  begin
    writehex((data[i] shr 12) and 15);
    writehex((data[i] shr 8) and 15);
    writehex((data[i] shr 4) and 15);
    writehex(data[i] and 15);
  end; {for}
  writeln(f,' ;');
end; {init16}

{write out one init parameter for 8-bit memory}
{m is the block memory number and n is the initialization string number}
procedure init8(m,n: integer; data: array of integer);
var
  i: integer;
begin
  write(f,'INST "pgmram',m,'" INIT_');
  writehex((n shr 4) and 15);
  writehex(n and 15);
  write(f,' = 256''h');
  for i := 31 downto 0 do
  begin
    writehex((data[i] shr 4) and 15);
    writehex(data[i] and 15);
  end; {for}
  writeln(f,' ;');
end; {init8}

{write out one init parameter for 4-bit memory}
{m is the block memory number and n is the initialization string number}
procedure init4(m,n: integer; data: array of integer);
var
  i: integer;
begin
  write(f,'INST "pgmram',m,'" INIT_');
  writehex((n shr 4) and 15);
  writehex(n and 15);
  write(f,' = 256''h');
  for i := 63 downto 0 do
  begin
    writehex(data[i] and 15);
  end; {for}
  writeln(f,' ;');
end; {init4}

{write out one init parameter for parity memory}
procedure initp(n: integer; data: array of integer);
var
  i,j,k: integer;
begin
  write(f,'   .INITP_');
  writehex((n shr 4) and 15);
  writehex(n and 15);
  write(f,'(256''h');
  for i := 31 downto 0 do
  begin
    k := 0;
    for j := 7 downto 0 do
    begin
      k := k shl 1;
      k := k or data[i*8+j];
    end; {for j}
    writehex((k shr 4) and 15);
    writehex(k and 15);
  end; {for i}
  writeln(f,'),');
end; {initp}

{pass 30 writes initialization file}
procedure pass30;
var
  i,j: integer;
begin
  {create file}
  AssignFile(f,fn+'.c16');
  Rewrite(f);
  if ioresult = 0 then
  begin
    {start symbol}
    write(f,'$');
    {initialize checksum and write data}
    j := 0;
    for i := 0 to paddr-1 do begin
      writeword(op0[i]);
      j := j + op0[i];
    end;
    {add checksum}
    j := j and 65535;
    writeword(-j);
    {end symbol}
    writeln(f,'#'); 
  end else begin
    writeln('ERROR creating binary output file!');
  end; {if}
  {finished}
  CloseFile(f);
end; {pass30}

{pass 31 writes 1 RAM initialization code into UCF file}
procedure pass31;
var
  i,j: integer;
  d: array [0..255] of integer;
begin
  {create file}
  AssignFile(f,fn+'.ucf');
  Rewrite(f);
  if ioresult = 0 then
  begin
    writeln(f,'# Program RAM initialization generated by CPU16 assembler v0.9');
    writeln(f,'# (C) Copyright 2009 John B. Stephensen  All rights reserved.');
    {initialize data bits}
    for i := 0 to 63 do
    begin
      for j := 0 to 15 do d[j] := op0[16*i+j] and 65535;
      init16(i,d);
    end; {for}
  end else begin
    writeln('ERROR creating UCF output file!');
  end; {if}
  {finished}
  CloseFile(f);
end; {pass31}

{pass 32 writes 2 RAM initialization code into UCF file}
procedure pass32;
var
  i,j: integer;
  d: array [0..255] of integer;
begin
  {create file}
  AssignFile(f,fn+'.ucf');
  Rewrite(f);
  if ioresult = 0 then
  begin
    writeln(f,'# Program RAM initialization generated by CPU16 assembler v0.9');
    writeln(f,'# (C) Copyright 2009 John B. Stephensen  All rights reserved.');
    {lower RAM}
    writeln(f,'# instruction LSB');
    {initialize bits 7-0}
    for i := 0 to 63 do
    begin
      for j := 0 to 31 do d[j] := op0[32*i+j] and 255;
      init8(0,i,d);
    end; {for}
    {upper RAM}
    writeln(f,'# instruction MSB');
    {initialize bits 16-9}
    for i := 0 to 63 do
    begin
      for j := 0 to 31 do d[j] := (op0[32*i+j] shr 8) and 255;
      init8(1,i,d);
    end; {for}
  end else begin
    writeln('ERROR creating UCF output file!');
  end; {if}
  {finished}
  CloseFile(f);
end; {pass32}

{pass 34 writes 4 RAM initialization code into UCF file}
procedure pass34;
var
  i,j: integer;
  d: array [0..255] of integer;
begin
  {create file}
  AssignFile(f,fn+'.ucf');
  Rewrite(f);
  if ioresult = 0 then
  begin
    writeln(f,'# Program RAM initialization generated by CPU16 assembler v0.9');
    writeln(f,'# (C) Copyright 2009 John B. Stephensen  All rights reserved.');
    {1st RAM}
    writeln(f,'# RAM 0');
    {initialize bits 3-0}
    for i := 0 to 63 do
    begin
      for j := 0 to 63 do d[j] := op0[64*i+j] and 15;
      init4(0,i,d);
    end; {for}
    {2nd RAM}
    writeln(f,'# RAM 1');
    {initialize bits 7-4}
    for i := 0 to 63 do
    begin
      for j := 0 to 63 do d[j] := (op0[64*i+j] shr 4) and 15;
      init4(1,i,d);
    end; {for}
    {3rd RAM}
    writeln(f,'# RAM 2');
    {initialize bits 3-0}
    for i := 0 to 63 do
    begin
      for j := 0 to 63 do d[j] := (op0[64*i+j] shr 8) and 15;
      init4(2,i,d);
    end; {for}
    {4th RAM}
    writeln(f,'# RAM 3');
    {initialize bits 15-12}
    for i := 0 to 63 do
    begin
      for j := 0 to 63 do d[j] := (op0[64*i+j] shr 12) and 15;
      init4(3,i,d);
    end; {for}
  end else begin
    writeln('ERROR creating UCF output file!');
  end; {if}
  {finished}
  CloseFile(f);
end; {pass34}

{main program}
begin
  {initialize}
  line := 0;
  errors := 0;
  lastsym := 0;
  paddr := 0;
  romsize := 2;
  maxaddr := 2047;
  writeln('CPU16A Assembler v0.9 - Generates UCF file to initialize block RAM.');
  {open source file}
  repeat
    write('Enter source file name: C:\');
    readln(fn);
    fn := 'C:\' + fn;
    AssignFile(f,fn+'.asm');
  until ioresult = 0;
  {parse source code and generate object code in memory}
  pass1;
  {generate listing}
  pass2;
  writeln(errors,' errors, ',lastsym,' symbols used, ',paddr,' instructions.');
  {generate program RAM file if no errors}
  if errors = 0 then
  begin
    case romsize of
      0: pass30;
      1: pass31;
      2: pass32;
      4: pass34;
    else
      writeln('Invalid number of memory blocks - no output file generated!');
    end;
  end;
  write('Press ENTER to exit...'); readln;
end.
